import attr
from typing import Optional

from ..utils.spectrogram import LogMelSpectrogram, torch
from .ts import TimeSeriesModality
from .traits import CalculableModalityTrait


@attr.define
class MelModality(CalculableModalityTrait, TimeSeriesModality):
    mel_fn: LogMelSpectrogram
    sample_rate: float = attr.field(init=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.sample_rate = self.mel_fn.rate

    def calculate(self):
        wav = self.base_modality.fetch()
        with torch.no_grad():
            wav = torch.from_numpy(wav)  # (t c)
            mel = self.mel_fn(wav, dim=0).numpy()  # (t c d)
        return mel

    @property
    def checksum(self):
        return str(self.mel_fn)



import math
import torch
import warnings
from torch import Tensor
from typing import Optional


def _hz_to_mel(freq: float, mel_scale: str = "htk") -> float:
    r"""Convert Hz to Mels.
    Args:
        freqs (float): Frequencies in Hz
        mel_scale (str, optional): Scale to use: ``htk`` or ``slaney``. (Default: ``htk``)
    Returns:
        mels (float): Frequency in Mels
    """

    if mel_scale not in ["slaney", "htk"]:
        raise ValueError('mel_scale should be one of "htk" or "slaney".')

    if mel_scale == "htk":
        return 2595.0 * math.log10(1.0 + (freq / 700.0))

    # Fill in the linear part
    f_min = 0.0
    f_sp = 200.0 / 3

    mels = (freq - f_min) / f_sp

    # Fill in the log-scale part
    min_log_hz = 1000.0
    min_log_mel = (min_log_hz - f_min) / f_sp
    logstep = math.log(6.4) / 27.0

    if freq >= min_log_hz:
        mels = min_log_mel + math.log(freq / min_log_hz) / logstep

    return mels


def _mel_to_hz(mels: Tensor, mel_scale: str = "htk") -> Tensor:
    """Convert mel bin numbers to frequencies.
    Args:
        mels (Tensor): Mel frequencies
        mel_scale (str, optional): Scale to use: ``htk`` or ``slaney``. (Default: ``htk``)
    Returns:
        freqs (Tensor): Mels converted in Hz
    """

    if mel_scale not in ["slaney", "htk"]:
        raise ValueError('mel_scale should be one of "htk" or "slaney".')

    if mel_scale == "htk":
        return 700.0 * (10.0 ** (mels / 2595.0) - 1.0)

    # Fill in the linear scale
    f_min = 0.0
    f_sp = 200.0 / 3
    freqs = f_min + f_sp * mels

    # And now the nonlinear scale
    min_log_hz = 1000.0
    min_log_mel = (min_log_hz - f_min) / f_sp
    logstep = math.log(6.4) / 27.0

    log_t = mels >= min_log_mel
    freqs[log_t] = min_log_hz * torch.exp(logstep * (mels[log_t] - min_log_mel))

    return freqs


def create_fb_matrix(
    n_freqs: int,
    f_min: float,
    f_max: float,
    n_mels: int,
    sample_rate: int,
    norm: Optional[str] = None,
    mel_scale: str = "htk",
) -> Tensor:
    r"""Create a frequency bin conversion matrix.
    Args:
        n_freqs (int): Number of frequencies to highlight/apply
        f_min (float): Minimum frequency (Hz)
        f_max (float): Maximum frequency (Hz)
        n_mels (int): Number of mel filterbanks
        sample_rate (int): Sample rate of the audio waveform
        norm (Optional[str]): If 'slaney', divide the triangular mel weights by the width of the mel band
        (area normalization). (Default: ``None``)
        mel_scale (str, optional): Scale to use: ``htk`` or ``slaney``. (Default: ``htk``)
    Returns:
        Tensor: Triangular filter banks (fb matrix) of size (``n_freqs``, ``n_mels``)
        meaning number of frequencies to highlight/apply to x the number of filterbanks.
        Each column is a filterbank so that assuming there is a matrix A of
        size (..., ``n_freqs``), the applied result would be
        ``A * create_fb_matrix(A.size(-1), ...)``.
    """

    if norm is not None and norm != "slaney":
        raise ValueError("norm must be one of None or 'slaney'")

    # freq bins
    # Equivalent filterbank construction by Librosa
    all_freqs = torch.linspace(0, sample_rate // 2, n_freqs)

    # calculate mel freq bins
    m_min = _hz_to_mel(f_min, mel_scale=mel_scale)
    m_max = _hz_to_mel(f_max, mel_scale=mel_scale)

    m_pts = torch.linspace(m_min, m_max, n_mels + 2)
    f_pts = _mel_to_hz(m_pts, mel_scale=mel_scale)

    # calculate the difference between each mel point and each stft freq point in hertz
    f_diff = f_pts[1:] - f_pts[:-1]  # (n_mels + 1)
    slopes = f_pts.unsqueeze(0) - all_freqs.unsqueeze(1)  # (n_freqs, n_mels + 2)
    # create overlapping triangles
    zero = torch.zeros(1)
    down_slopes = (-1.0 * slopes[:, :-2]) / f_diff[:-1]  # (n_freqs, n_mels)
    up_slopes = slopes[:, 2:] / f_diff[1:]  # (n_freqs, n_mels)
    fb = torch.max(zero, torch.min(down_slopes, up_slopes))

    if norm is not None and norm == "slaney":
        # Slaney-style mel is scaled to be approx constant energy per channel
        enorm = 2.0 / (f_pts[2 : n_mels + 2] - f_pts[:n_mels])
        fb *= enorm.unsqueeze(0)

    if (fb.max(dim=0).values == 0.0).any():
        warnings.warn(
            "At least one mel filterbank has all zero values. "
            f"The value for `n_mels` ({n_mels}) may be set too high. "
            f"Or, the value for `n_freqs` ({n_freqs}) may be set too low."
        )

    return fb



from .sample import MultimodalSample


class MultimodalDataset:
    Sample = MultimodalSample

    def __init__(self, ids, modality_factories):
        super().__init__()
        self.ids = ids
        self.modality_factories = modality_factories
        self.samples = self.create_samples()

    def create_samples(self):
        samples = [self.Sample(id) for id in self.ids]
        for sample in samples:
            for modality_factory in self.modality_factories:
                sample.add_modality(modality_factory)
        return samples

    def __getitem__(self, index):
        return self.samples[index].fetch()

    def __len__(self):
        return len(self.samples)



import numpy as np
from dataclasses import dataclass

try:
    import torch
    from torchaudio.transforms import MelSpectrogram as _MelSpectrogramBase
except:
    raise ImportError(
        "To use the mel spectrogram feature, PyTorch and Torchaudio is required. "
        "Install it with: pip install torch torchaudio"
    )

from .torchaudio_future import create_fb_matrix


def lws_hann(n):
    """
    symmetric hann from lws, which start from 1 instead of 0
    """
    return 0.5 * (1 - np.cos(2 * np.pi * (np.arange(1, 2 * n, 2)) / (2 * n)))


def synthwin(awin, fshift, swin=None):
    """
    synthwin from lws
    """
    # returns the normalized synthesis window for perfect reconstruction
    fsize = len(awin)
    Q = int(np.ceil(fsize * 1.0 / fshift))
    if swin is None:
        swin = awin
    twin = awin * swin
    tmp_fsize = Q * fshift

    w = np.hstack([twin, np.zeros((tmp_fsize - fsize,))])
    w = np.sum(np.reshape(w, (Q, fshift)), axis=0)
    w = np.tile(w, (1, Q))[0, :fsize]

    if min(w) <= 0:
        raise ValueError("The normalizer is not strictly positive")

    swin = swin / w

    return swin


@dataclass(eq=False)
class LogMelSpectrogram(_MelSpectrogramBase):
    sample_rate: int = 16000
    n_mels: int = 80
    f_min: int = 125
    f_max: int = 7600
    hop_length: int = 256
    legacy: bool = True

    def __post_init__(self):
        win_length = 4 * self.hop_length
        n_fft = 2 ** (win_length - 1).bit_length()

        window_fn = torch.hann_window
        if self.legacy:
            window_fn = self.lws_window_fn

        super().__init__(
            self.sample_rate,
            n_fft,
            win_length,
            self.hop_length,
            self.f_min,
            self.f_max,
            n_mels=self.n_mels,
            pad=0,
            window_fn=window_fn,
            power=1,
            normalized=False,
            norm="slaney",
            # mel_scale="slaney",
        )

        self.mel_scale.register_buffer(
            "fb",
            create_fb_matrix(
                self.n_fft // 2 + 1,
                self.f_min,
                self.f_max,
                self.n_mels,
                self.sample_rate,
                "slaney",
                "slaney",
            ),
        )

    def lws_window_fn(self, win_length):
        awin = torch.from_numpy(lws_hann(win_length)).double().sqrt()
        awin = (awin * synthwin(awin, self.hop_length)).sqrt()
        return awin.float()

    @property
    def rate(self):
        return self.sample_rate / self.hop_length

    @property
    def min_level_db(self):
        if self.legacy:
            return -100
        return -200  # eps = 1e-10

    @property
    def ref_level_db(self):
        assert self.legacy
        return 20

    @property
    def min_level(self):
        # 10 ^ (db / 20)
        return np.exp(self.min_level_db / 20 * np.log(10))

    def forward(self, wav, dim=-1, drop_last=True):
        """
        Args:
            wav: (... t ...)
            dim: the dim for t
        Returns:
            mel: (... t' ... c)
        """
        if not isinstance(wav, torch.Tensor):
            wav = torch.tensor(wav)

        # swap t to the last dim
        wav = wav.transpose(dim, -1)
        mel = super().forward(wav)
        mel = mel.clamp_min(self.min_level).log10()

        if self.legacy:
            # the legacy wavenet normalization
            # https://github.com/r9y9/wavenet_vocoder/blob/42a488b74b901db3fdf49689d9d8503fdc109c11/audio.py
            mel = 20 * mel - self.ref_level_db
            mel = ((mel - self.min_level_db) / -self.min_level_db).clamp(0, 1)
        else:
            # the new wavenet (espnet), which does not do normalization
            # https://github.com/r9y9/wavenet_vocoder/blob/c93a556466a3378be8f67cd3a0e9d689915c4fab/audio.py
            pass

        if drop_last:
            mel = mel[..., :-1]

        # swap t back
        mel = mel.transpose(dim, -1)
        # swap mel as the last dim
        mel = mel.transpose(-1, -2)

        return mel


if __name__ == "__main__":
    mel_fn = LogMelSpectrogram()
    print(mel_fn)

    import sys
    import librosa
    from pathlib import Path

    path = Path(sys.argv[1])
    print(path)
    wav = librosa.load(path, 16000)[0]
    print(wav.shape)
    mel = mel_fn(wav)
    print(mel)
    print(mel.shape)

    mel2 = torch.load(f"../wavenet_vocoder/template.pth")
    mel2 = torch.from_numpy(mel2)
    print(mel2)
    print(mel2.shape)
    print(torch.isclose(mel[2:-1], mel2[:-1], 1e-4).all())



import attr
from functools import partial


@attr.define
class MultimodalSample:
    id: str
    modalities: list = attr.field(init=False, factory=list, repr=False)

    def add_modality(self, factory: partial):
        modality = factory(sample=self)
        if modality.name in map(lambda m: m.name, self.modalities):
            raise ValueError(
                "Modality should have unique name, "
                f"but got more than 1: {modality.name}."
            )
        self.modalities.append(modality)

    def fetch(self):
        for modality in self.modalities:
            try:
                modality.load()
            except Exception as e:
                raise RuntimeError(f"Load {modality} failed.") from e

        info = self.generate_info()

        data = dict(info=info)

        for modality in self.modalities:
            try:
                data[modality.name] = modality.fetch(info=info)
            except Exception as e:
                raise RuntimeError(f"Fetch {modality} failed.") from e

        data = {
            modality.name: modality.fetch(info=info) for modality in self.modalities
        }

        data["info"] = info

        for modality in self.modalities:
            modality.unload()

        return data

    def get_modality_by_name(self, name):
        return next(modality for modality in self.modalities if modality.name == name)

    def generate_info(self) -> dict:
        return {}



import attr
from pathlib import Path
from functools import partial
from typing import Optional, Any

from ..sample import MultimodalSample


@attr.define
class Modality:
    """
    Args:
        name: the name of the modality.
        root: the data root of this modality.
        suffix: the suffix of the data, e.g. .mp4 or /*.jpg, yielding path: {root}/{id}.mp4 or paths: {root}/{id}/*.jpg.
        cache: a dict for caching which may be managed by a proxy process.
    """

    sample: MultimodalSample
    name: str = attr.field(converter=str)
    root: Path = attr.field(converter=Path)
    suffix: str = attr.field(converter=str)
    cache: Optional[dict] = attr.field(kw_only=True, default=None)

    _paths: list[Path] = attr.field(init=False, repr=False, default=[])
    _path: Optional[Path] = attr.field(init=False, repr=False, default=None)
    _loaded: Any = attr.field(init=False, repr=False, default=None)

    def __attrs_post_init__(self):
        if "*" in self.suffix:
            paths = (self.root / self.sample.id).glob(self.suffix)
            self._paths = sorted(paths)
        else:
            self._path = self.root / (self.sample.id + self.suffix)

    @classmethod
    def create_factory(cls, **kwargs):
        return partial(cls, **kwargs)

    @property
    def path(self) -> Optional[Path]:
        return self._path

    @property
    def paths(self) -> list[Path]:
        return self._paths

    @property
    def loaded(self):
        return self._loaded

    @property
    def cached(self):
        if self.cache is not None and self._cache_key in self.cache:
            return self.cache[self._cache_key]

    def load(self):
        if self.loaded is None:
            if (cached := self.cached) is None:
                self._loaded = self._load_impl()
                if self.cache is not None:
                    self.cache[self._cache_key] = self.loaded
            else:
                self._loaded = cached

    def fetch(self, info={}):
        self.load()
        return self._fetch_impl(info=info)

    def unload(self):
        self._loaded = None

    @property
    def _cache_key(self):
        return (self.sample.id, self.name)

    def _load_impl(self):
        pass

    def _fetch_impl(self, info={}):
        return self.loaded



# https://github.com/NVIDIA/mellotron/blob/master/yin.py
# adapted from https://github.com/patriceguyot/Yin

import numpy as np


def differenceFunction(x, N, tau_max):
    """
    Compute difference function of data x. This corresponds to equation (6) in [1]
    This solution is implemented directly with Numpy fft.


    :param x: audio data
    :param N: length of data
    :param tau_max: integration window size
    :return: difference function
    :rtype: list
    """

    x = np.array(x, np.float64)
    w = x.size
    tau_max = min(tau_max, w)
    x_cumsum = np.concatenate((np.array([0.0]), (x * x).cumsum()))
    size = w + tau_max
    p2 = (size // 32).bit_length()
    nice_numbers = (16, 18, 20, 24, 25, 27, 30, 32)
    size_pad = min(x * 2 ** p2 for x in nice_numbers if x * 2 ** p2 >= size)
    fc = np.fft.rfft(x, size_pad)
    conv = np.fft.irfft(fc * fc.conjugate())[:tau_max]
    return x_cumsum[w : w - tau_max : -1] + x_cumsum[w] - x_cumsum[:tau_max] - 2 * conv


def cumulativeMeanNormalizedDifferenceFunction(df, N):
    """
    Compute cumulative mean normalized difference function (CMND).

    This corresponds to equation (8) in [1]

    :param df: Difference function
    :param N: length of data
    :return: cumulative mean normalized difference function
    :rtype: list
    """

    cmndf = df[1:] * range(1, N) / np.cumsum(df[1:]).astype(float)  # scipy method
    return np.insert(cmndf, 0, 1)


def getPitch(cmdf, tau_min, tau_max, harmo_th=0.1):
    """
    Return fundamental period of a frame based on CMND function.

    :param cmdf: Cumulative Mean Normalized Difference function
    :param tau_min: minimum period for speech
    :param tau_max: maximum period for speech
    :param harmo_th: harmonicity threshold to determine if it is necessary to compute pitch frequency
    :return: fundamental period if there is values under threshold, 0 otherwise
    :rtype: float
    """
    tau = tau_min
    while tau < tau_max:
        if cmdf[tau] < harmo_th:
            while tau + 1 < tau_max and cmdf[tau + 1] < cmdf[tau]:
                tau += 1
            return tau
        tau += 1

    return 0  # if unvoiced


def compute_yin(
    sig, sr, w_len=512, w_step=256, f0_min=100, f0_max=500, harmo_thresh=0.1
):
    """

    Compute the Yin Algorithm. Return fundamental frequency and harmonic rate.

    :param sig: Audio signal (list of float)
    :param sr: sampling rate (int)
    :param w_len: size of the analysis window (samples)
    :param w_step: size of the lag between two consecutives windows (samples)
    :param f0_min: Minimum fundamental frequency that can be detected (hertz)
    :param f0_max: Maximum fundamental frequency that can be detected (hertz)
    :param harmo_tresh: Threshold of detection. The yalgorithmù return the first minimum of the CMND function below this treshold.

    :returns:

        * pitches: list of fundamental frequencies,
        * harmonic_rates: list of harmonic rate values for each fundamental frequency value (= confidence value)
        * argmins: minimums of the Cumulative Mean Normalized DifferenceFunction
        * times: list of time of each estimation
    :rtype: tuple
    """

    tau_min = int(sr / f0_max)
    tau_max = int(sr / f0_min)

    timeScale = range(
        0, len(sig) - w_len, w_step
    )  # time values for each analysis window
    times = [t / float(sr) for t in timeScale]
    frames = [sig[t : t + w_len] for t in timeScale]

    pitches = [0.0] * len(timeScale)
    harmonic_rates = [0.0] * len(timeScale)
    argmins = [0.0] * len(timeScale)

    for i, frame in enumerate(frames):
        # Compute YIN
        df = differenceFunction(frame, w_len, tau_max)
        cmdf = cumulativeMeanNormalizedDifferenceFunction(df, tau_max)
        p = getPitch(cmdf, tau_min, tau_max, harmo_thresh)

        # Get results
        if np.argmin(cmdf) > tau_min:
            argmins[i] = float(sr / np.argmin(cmdf))
        if p != 0:  # A pitch was found
            pitches[i] = float(sr / p)
            harmonic_rates[i] = cmdf[p]
        else:  # No pitch, but we compute a value of the harmonic rate
            harmonic_rates[i] = min(cmdf)

    return pitches, harmonic_rates, argmins, times



import attr
import logging
import numpy as np

from .modality import Modality

logger = logging.getLogger(__name__)


@attr.define
class TimeSeriesModality(Modality):
    sample_rate: float

    @property
    def duration(self):
        assert self.loaded is not None
        return len(self.loaded) / self.sample_rate

    def _slice(self, x, t0, t1):
        start = None if t0 is None else int(t0 * self.sample_rate)
        stop = None if t1 is None else int(t1 * self.sample_rate)
        if stop is not None:
            x = self._ensure_length(x, stop)
        return x[start:stop]

    def _ensure_length(self, x, n):
        if n > len(x):
            diff = len(x) - n

            logger.warning(
                f"Modality {self.name} of sample {self.sample} does not have enough time-steps. "
                f"Expect at least {n} but got {len(x)}. "
                f"{diff} time-steps are padded."
            )

            x = self._pad_fn(x, diff)

        return x

    @staticmethod
    def _pad_fn(x, n):
        return np.pad(x, (0, n))

    def _fetch_impl(self, info={}):
        assert self.loaded is not None
        return self._slice(self.loaded, info.get("t0", None), info.get("t1", None))



import attr
import logging
from PIL import Image
from typing import Callable

from .ts import TimeSeriesModality

logger = logging.getLogger(__name__)

dumb_image = Image.new("RGB", (32, 32))


@attr.define
class RgbsModality(TimeSeriesModality):
    """A rgb sequence modality for video."""

    transform: Callable
    aggragate: Callable

    @property
    def duration(self):
        return len(self.paths) / self.sample_rate

    def _fetch_impl(self, *, info={}):
        paths = self._slice(self.paths, info.get("t0", None), info.get("t1", None))
        frames = list(map(self.transform, map(self._load_pil, paths)))
        return self.aggragate(frames)

    @staticmethod
    def _pad_fn(x, n):
        return x + [None] * n

    @staticmethod
    def _load_pil(path):
        if path is None:
            return dumb_image

        try:
            image = Image.open(path)
        except:
            logger.warning(f"Open {path} failed, use an empty picture instead.")
            image = dumb_image

        return image
